"""
Manages dependencies, resolving them into tiers,
where items within each tier can run in parallel,
but each tier must execute before subsequent tiers.
"""

from typing import Any, Callable, Iterable, List, Set


class DependencyResolutionError(ValueError):
    """Raised when a dependency cannot be resolved."""
    pass


class DependencyLoopError(DependencyResolutionError):
    """Raised when a dependency loop is detected."""
    def __init__(self, *items: List['Dep']):
        super().__init__('Loop detected between {!r}'.format(*items))
        self.items = items
    

class Dep(object):
    """Base class for objects which have dependencies."""

    def __init__(self):
        self._deps: Set[Dep] = set()

    def add_dependency(self, dep: 'Dep'):
        """Add a dependency."""
        self._deps.add(dep)

    def add_dependencies(self, deps: 'Iterable[Dep]'):
        """Add multiple dependencies."""
        self._deps.update(deps)

    def depends_on(self, dep: 'Dep') -> bool:
        """Return True if this object depends on the given object."""
        return dep in self._deps

    def get_dependencies(self):
        """Return the dependencies."""
        return self._deps


def resolve_deps(deps: Set[Dep]) -> List[Set[Dep]]:
    """Resolve dependencies into tiers which only depend upon prior tiers.

    :param deps: The dependencies to resolve.
    :return: A list of sets of dependencies, where each set depends only on
             dependencies in prior sets.  i.e., the sequence of sets must
             be executed in order, but each set within it can be divided
            into parallel tasks.
    """
    assert isinstance(deps, set), "resolve_deps expects a set"

    # build the set of all dependencies from deps and their dependencies
    all_deps = set(deps) | set(d for dep in deps for d in dep.get_dependencies())

    # sort the deps by their dependencies
    unresolved = set(all_deps)

    tiers = []
    resolved = set()

    while len(unresolved) > 0:
        tier = set()

        tier = set(d for d in unresolved if all(dep in resolved for dep in d.get_dependencies()))
        if len(tier) == 0:
            raise DependencyLoopError(*unresolved)

        tiers.append(tier)

        unresolved -= tier
        resolved |= tier

    return tiers



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['paradeps']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['deps = deps.__main__:main']}

setup_kwargs = {
    'name': 'paradeps',
    'version': '0.1.0',
    'description': 'Resolves dependencies into tiers suitable for parallel work',
    'long_description': '# paradeps\n\n## This resolves dependencies into parallelizable execution tiers.  For example, given the inputs:\n\n    a, b, c, d, f\n\nWhere a and b depend on f, this will return:\n\n    [{c,d,f}, {a,b}]\n\n\n## Usage\n\n    from paradeps import Dependent\n\n    class Item(Dep):\n       ...\n\n    def process_items_in_parallel(items):\n       ...\n\n    items = [Item() for i in range(5)]\n\n    items[0].add_dependency(items[2])\n    items[2].add_dependency(items[4])\n\n    tiers = resolve_deps(items)\n\n    for tier in tiers:\n        process_items_in_parallel(tier)\n\n\n## License\n\nReleased under the terms of Aferro GPL version 3.0 only.  See `LICENSE.txt` for further details.\n\n\n## Authors / Contact\n\nEmail mailto::leebraid@gmail.com[mailto::leebraid@gmail.com].\n',
    'author': 'Lee Braiden',
    'author_email': 'leebraid@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
